import pytest
from django.test import override_settings
from atol.models import Receipt, ReceiptStatus

pytestmark = pytest.mark.django_db(transaction=True)


def test_receipt_302_normal_redirect(client):
    data = {
        'callback_url': '',
        'daemon_code': 'prod-agent-3',
        'device_code': 'KSR13.11-8-18',
        'error': None,
        'group_code': 'mybook-ru_1815',
        'payload': {
            'ecr_registration_number': '0000932756018558',
            'fiscal_document_attribute': 4146968358,
            'fiscal_document_number': 40,
            'fiscal_receipt_number': 1,
            'fn_number': '8710000100942521',
            'fns_site': 'www.nalog.ru',
            'receipt_datetime': '26.07.2017 10:32:00',
            'shift_number': 19,
            'total': 12
        },
        'status': 'done',
        'timestamp': '26.07.2017 10:32:21',
        'uuid': 'd407f2bf-edb8-43c9-aac4-468c05f1a8d8'
    }
    receipt = Receipt.objects.create(content=data, status=ReceiptStatus.received)

    resp = client.get(receipt.ofd_link)
    assert resp.status_code == 302
    assert resp['Location'].startswith('https://lk.platformaofd')
    assert receipt.content['payload']['fn_number'] in resp['Location']

    with override_settings(RECEIPTS_OFD_URL_TEMPLATE=u'fake?t={t}&s={s}&fn={fn}&i={fd}&fp={fp}&n={n}'):
        resp = client.get(receipt.ofd_link)
        assert resp.status_code == 302
        assert '20170726T103200' in resp['Location']


@pytest.mark.parametrize('content', [
    ['Dummy'],
    {'a': 'b'},
    None,
])
def test_receipt_404(content, client):
    receipt = Receipt.objects.create(content=content)
    response = client.get(receipt.ofd_link, expect_errors=True)
    assert response.status_code == 404



import math
import logging
from uuid import uuid4
from datetime import timedelta

from django.db import transaction
from django.utils import timezone
from django.apps import apps
from celery.exceptions import MaxRetriesExceededError
from celery import shared_task

from atol.core import AtolAPI
from atol.models import ReceiptStatus
from atol.exceptions import AtolUnrecoverableError, NoEmailAndPhoneError, AtolReceiptNotProcessed

logger = logging.getLogger(__name__)


@shared_task(name='atol_create_receipt', bind=True, max_retries=4, time_limit=60, soft_time_limit=45)
def atol_create_receipt(self, receipt_id):
    """
    Change receipt status and the change date accordingly
    If received an unrecoverable error, stop any further attempts to init a receipt and mark its status as failed
    """
    atol = AtolAPI()
    Receipt = apps.get_model('atol', 'Receipt')
    receipt = Receipt.objects.get(id=receipt_id)

    try:
        params = receipt.get_params()
    except NoEmailAndPhoneError:
        # this email should have been sent, but we got neither email
        logger.warning('unable to init receipt %s due to missing email/phone', receipt.id)
        receipt.declare_failed(status=ReceiptStatus.no_email_phone)
        return

    if receipt.status not in [ReceiptStatus.created, ReceiptStatus.retried]:
        logger.error('receipt %s has invalid status: %s', receipt.uuid, receipt.status)
        return

    try:
        receipt_data = atol.sell(**params)
    except AtolUnrecoverableError as exc:
        logger.error('unable to init receipt %s with params %s due to %s', receipt.id, params, exc,
                     exc_info=True, extra={'data': {'payment_params': params}})
        receipt.declare_failed()
    except Exception as exc:
        logger.warning('failed to init receipt %s with params %s due to %s', receipt.id, params, exc,
                       exc_info=True, extra={'data': {'payment_params': params}})
        try:
            countdown = 60 * int(math.exp(self.request.retries))
            logger.info('retrying to create receipt %s with params %s countdown %s due to %s',
                        receipt.id, params, countdown, exc)
            self.retry(countdown=countdown)
        except MaxRetriesExceededError:
            logger.error('run out of attempts to create receipt %s with params %s due to %s',
                         receipt.id, params, exc)
            receipt.declare_failed()
    else:
        with transaction.atomic():
            receipt.initiate(uuid=receipt_data.uuid)
            transaction.on_commit(
                lambda: atol_receive_receipt_report.apply_async(args=(receipt.id,), countdown=60)
            )


@shared_task(name='atol_receive_receipt_report', bind=True, max_retries=4, time_limit=60, soft_time_limit=45)
def atol_receive_receipt_report(self, receipt_id):
    """
    Attempt to retrieve a receipt report for given receipt_id
    If received an unrecoverable error, then stop any further attempts to receive the report
    """
    atol = AtolAPI()
    Receipt = apps.get_model('atol', 'Receipt')
    receipt = Receipt.objects.get(id=receipt_id)

    if not receipt.uuid:
        logger.error('receipt %s does not have a uuid', receipt.id)
        return

    if receipt.status not in [ReceiptStatus.initiated, ReceiptStatus.retried]:
        logger.error('receipt %s has invalid status: %s', receipt.uuid, receipt.status)
        return

    try:
        report = atol.report(receipt.uuid)
    except AtolUnrecoverableError as exc:
        logger.error('unable to fetch report for receipt %s due to %s',
                     receipt.id, exc, exc_info=True)
        receipt.declare_failed()
    except AtolReceiptNotProcessed as exc:
        logger.warning('unable to fetch report for receipt %s due to %s',
                       receipt.id, exc, exc_info=True)
        logger.info('repeat receipt registration: id %s; old internal_uuid %s',
                    receipt.id, receipt.internal_uuid)
        with transaction.atomic():
            receipt.internal_uuid = uuid4()
            receipt.status = ReceiptStatus.retried
            receipt.save(update_fields=['internal_uuid', 'status'])
            transaction.on_commit(
                lambda: atol_create_receipt.apply_async(args=(receipt.id,), countdown=60)
            )
    except Exception as exc:
        logger.warning('failed to fetch report for receipt %s due to %s',
                       receipt.id, exc, exc_info=True)
        try:
            countdown = 60 * int(math.exp(self.request.retries))
            logger.info('retrying to receive receipt %s with countdown %s due to %s',
                        receipt.id, countdown, exc)
            self.retry(countdown=countdown)
        except MaxRetriesExceededError:
            logger.error('run out of attempts to create receipt %s due to %s',
                         receipt.id, exc)
            receipt.declare_failed()
    else:
        with transaction.atomic():
            receipt.receive(content=report.data)


@shared_task(name='atol_retry_created_receipts', time_limit=3600)
def atol_retry_created_receipts():
    """
    Retry older receipt that have not been initiated by this time for some reason
    """
    Receipt = apps.get_model('atol', 'Receipt')
    now = timezone.now()

    created_receipts = (Receipt.objects
                        .filter(status=ReceiptStatus.created,
                                created_at__range=(now - timedelta(days=1), now)))

    logger.info('there are %s receipts waiting to be initiated', created_receipts.count())

    for receipt in created_receipts.only('pk').iterator():
        logger.info('retrying created receipt %s', receipt.id)
        atol_create_receipt.delay(receipt.id)


@shared_task(name='atol_retry_initiated_receipts', time_limit=3600)
def atol_retry_initiated_receipts():
    """
    Retry older receipts that have been initiated but have not been received receipt report from atol
    """
    Receipt = apps.get_model('atol', 'Receipt')
    now = timezone.now()

    initiated_receipts = (Receipt.objects
                          .filter(status=ReceiptStatus.initiated,
                                  initiated_at__range=(now - timedelta(days=1), now)))

    logger.info('there are %s initiated receipts waiting for report', initiated_receipts.count())

    for receipt in initiated_receipts.only('pk').iterator():
        logger.info('retrying initiated payment %s', receipt.id)
        atol_receive_receipt_report.delay(receipt.id)



import logging
from uuid import uuid4

import shortuuid
from django.contrib.postgres.fields import JSONField
from django.db import models, transaction
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse
from model_utils import Choices

from atol.signals import receipt_failed, receipt_initiated, receipt_received
from atol.exceptions import NoEmailAndPhoneError

logger = logging.getLogger(__name__)


ReceiptStatus = Choices(
    ('created', _('Ожидает инициации в системе оператора')),
    ('initiated', _('Иницирован в системе оператора')),
    ('retried', _('Повторно иницирован в системе оператора')),
    ('received', _('Получен от оператора')),
    ('no_email_phone', _('Отсутствует email/phone')),
    ('failed', _('Ошибка')),
)


class Receipt(models.Model):
    internal_uuid = models.UUIDField(default=uuid4, unique=True, editable=False)

    created_at = models.DateTimeField(_('Дата создания чека'), auto_now_add=True, editable=False)
    initiated_at = models.DateTimeField(_('Дата инициализации чека в системе оператора'), blank=True, null=True)
    retried_at = models.DateTimeField(_('Дата повторной инициализации чека в системе оператора'),
                                      blank=True, null=True)
    received_at = models.DateTimeField(_('Дата получения чека от оператора'), blank=True, null=True)
    failed_at = models.DateTimeField(_('Дата ошибки'), blank=True, null=True)

    status = models.CharField(_('Статус чека'), max_length=16, choices=ReceiptStatus,
                              default=ReceiptStatus.created)
    uuid = models.TextField(_('Идентификатор чека'), null=True, editable=False,
                            help_text=_('Идентификатор чека платежа в системе оператора'))
    content = JSONField(_('Содержимое чека'), null=True, editable=False)

    user_email = models.CharField(_('Email пользователя'), max_length=254, null=True)
    user_phone = models.CharField(_('Телефон пользователя'), max_length=32, null=True)
    purchase_price = models.DecimalField(_('Цена покупки'), max_digits=8, decimal_places=2, null=True)
    purchase_name = models.TextField(_('Наименование покупки'), null=True)

    class Meta:
        verbose_name = _('Чек Атола')
        verbose_name_plural = _('Чеки Атола')
        ordering = ['id']

    @cached_property
    def ofd_link(self):
        """Return the receipt url"""
        return reverse('receipt', kwargs={'short_uuid': shortuuid.encode(self.internal_uuid)})

    @transaction.atomic()
    def declare_failed(self, status=None):
        logger.warning('declaring receipt %s as failed', self.id)
        self.status = status or ReceiptStatus.failed
        self.failed_at = timezone.now()
        self.save(update_fields=['status', 'failed_at'])
        receipt_failed.send(sender=None, receipt=self)

    def initiate(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        update_fields = list(kwargs.keys())

        now = timezone.now()
        if self.status == ReceiptStatus.retried:
            self.retried_at = now
            update_fields += ['retried_at']
        else:
            self.initiated_at = now
            self.status = ReceiptStatus.initiated
            update_fields += ['initiated_at', 'status']

        self.save(update_fields=update_fields)
        receipt_initiated.send(sender=None, receipt=self)

    def receive(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.status = ReceiptStatus.received
        self.received_at = timezone.now()
        self.save(update_fields=list(kwargs.keys()) + ['status', 'received_at'])
        receipt_received.send(sender=None, receipt=self)

    def get_params(self):
        params = {
            'timestamp': self.created_at.isoformat(),
            'transaction_uuid': str(self.internal_uuid),
            'purchase_price': float(self.purchase_price),
            'purchase_name': self.purchase_name or 'Оплата подписки'
        }
        if self.user_email:
            params['user_email'] = self.user_email
        elif self.user_phone:
            params['user_phone'] = self.user_phone  # don't strip leading plus nor non-russian country code
        else:
            raise NoEmailAndPhoneError

        return params



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-12-13 09:12
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('atol', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='receipt',
            name='retried_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Дата повторной инициализации чека в системе оператора'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='status',
            field=models.CharField(choices=[('created', 'Ожидает инициации в системе оператора'), ('initiated', 'Иницирован в системе оператора'), ('retried', 'Повторно иницирован в системе оператора'), ('received', 'Получен от оператора'), ('no_email_phone', 'Отсутствует email/phone'), ('failed', 'Ошибка')], default='created', max_length=16, verbose_name='Статус чека'),
        ),
    ]



from uuid import uuid4
from freezegun import freeze_time
import responses
import datetime
import mock
import pytest

from django.utils import timezone

from atol.core import AtolAPI, NewReceipt
from atol.models import Receipt, ReceiptStatus
from atol.tasks import (atol_create_receipt, atol_receive_receipt_report,
                        atol_retry_created_receipts, atol_retry_initiated_receipts)
from tests import ATOL_BASE_URL

pytestmark = pytest.mark.django_db(transaction=True)


@responses.activate
def test_created_receipt_ok():
    now = timezone.now()
    responses.add(responses.POST, ATOL_BASE_URL + '/getToken', status=200,
                  json={'code': 0, 'token': 'foobar'})

    receipt = Receipt.objects.create(user_email='foo@bar.com', purchase_price=707.1)

    with mock.patch.object(AtolAPI, 'sell', wraps=AtolAPI.sell) as sell_mock:
        with mock.patch.object(atol_receive_receipt_report, 'apply_async') as task_mock:
            sell_mock.return_value = NewReceipt(uuid='5869a6d9-1540-4ebb-a2a2-f1d11501f213', data=None)
            atol_create_receipt(receipt.id)
            assert len(task_mock.mock_calls) == 1

    receipt.refresh_from_db()
    assert receipt.uuid == '5869a6d9-1540-4ebb-a2a2-f1d11501f213'
    assert receipt.status == ReceiptStatus.initiated
    assert receipt.initiated_at > now


@responses.activate
def test_atol_create_failing_receipt_progressive_countdown():
    responses.add(responses.POST, ATOL_BASE_URL + '/getToken', status=200,
                  json={'code': 0, 'token': 'foobar'})
    responses.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=500)

    receipt = Receipt.objects.create(user_email='foo@bar.com', purchase_price=999)

    with mock.patch.object(atol_create_receipt, 'retry', wraps=atol_create_receipt.retry) as task_mock:
        atol_create_receipt.delay(receipt.id)
        assert len(task_mock.mock_calls) == 5
        assert [call[1]['countdown'] for call in task_mock.call_args_list] == [60, 120, 420, 1200, 3240]

    receipt.refresh_from_db()
    assert receipt.status == 'failed'


@responses.activate
def test_atol_create_receipt_stopped_on_unrecoverable_error():
    responses.add(responses.POST, ATOL_BASE_URL + '/getToken',
                  status=200, json={'code': 0, 'token': 'foobar'})
    responses.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell',
                  status=400, json={'error': {'code': 3}})

    receipt = Receipt.objects.create(user_email='foo@bar.com', purchase_price=999)
    atol_create_receipt(receipt.id)
    receipt.refresh_from_db()
    assert receipt.status == 'failed'


@pytest.mark.parametrize(['receipt_data', 'status'], [
    ({'status': 'failed', 'purchase_price': 299, 'user_email': 'foo@bar.com'}, ReceiptStatus.failed),
    ({'purchase_price': 299}, ReceiptStatus.no_email_phone),
])
def test_atol_create_receipt_fail(receipt_data, status):
    receipt = Receipt.objects.create(**receipt_data)

    with mock.patch.object(AtolAPI, 'sell') as sell_mock:
        atol_create_receipt(receipt.id)
        assert len(sell_mock.mock_calls) == 0

    receipt.refresh_from_db()
    assert receipt.status == status


@responses.activate
def test_atol_failing_receive_report_progressive_countdown():
    uuid = str(uuid4())
    receipt = Receipt.objects.create(status='initiated', uuid=uuid)

    responses.add(responses.POST, ATOL_BASE_URL + '/getToken',
                  status=200, json={'code': 0, 'token': 'foobar'})
    responses.add(responses.GET, ATOL_BASE_URL + '/ATOL-ProdTest-1/report/%s' % uuid,
                  status=500)

    with mock.patch.object(atol_receive_receipt_report, 'retry', wraps=atol_receive_receipt_report.retry) as task_mock:
        atol_receive_receipt_report.delay(receipt.id)
        assert len(task_mock.mock_calls) == 5
        assert [call[1]['countdown'] for call in task_mock.call_args_list] == [60, 120, 420, 1200, 3240]

    receipt.refresh_from_db()
    assert receipt.status == 'failed'


@responses.activate
def test_atol_fetch_receipt_report_stopped_on_unrecoverable_error():
    uuid = str(uuid4())
    receipt = Receipt.objects.create(status='initiated', uuid=uuid)

    responses.add(responses.POST, ATOL_BASE_URL + '/getToken',
                  status=200, json={'code': 0, 'token': 'foobar'})
    responses.add(responses.GET, ATOL_BASE_URL + '/ATOL-ProdTest-1/report/%s' % uuid,
                  status=400, json={'error': {'code': 3}})

    atol_receive_receipt_report(receipt.id)
    receipt.refresh_from_db()
    assert receipt.status == 'failed'


@responses.activate
def test_atol_fetch_receipt_report_check_receipt_status():
    uuid = str(uuid4())
    receipt = Receipt.objects.create(uuid=uuid)

    with mock.patch.object(AtolAPI, 'report') as report_mock:
        atol_receive_receipt_report(receipt.id)
        assert len(report_mock.mock_calls) == 0


@responses.activate
def test_retry_created_receipt_for_not_processed_receipt():
    uuid = str(uuid4())
    now = timezone.now()
    receipt = Receipt.objects.create(status='initiated', uuid=uuid, purchase_price=1234.5,
                                     user_phone='+79991234567')

    responses.add(responses.POST, ATOL_BASE_URL + '/getToken',
                  status=200, json={'code': 0, 'token': 'foobar'})
    responses.add(responses.GET, ATOL_BASE_URL + '/ATOL-ProdTest-1/report/%s' % uuid,
                  status=400, json={'error': {'code': 1}})

    with mock.patch.object(AtolAPI, 'sell', wraps=AtolAPI.sell) as sell_mock:
        sell_mock.return_value = NewReceipt(uuid='42ee0a7a-2b30-42f1-951f-1c131e2ab322', data=None)

        with mock.patch.object(atol_receive_receipt_report, 'apply_async'):
            atol_receive_receipt_report(receipt.id)
        assert len(sell_mock.mock_calls) == 1

    receipt.refresh_from_db()
    assert receipt.uuid == '42ee0a7a-2b30-42f1-951f-1c131e2ab322'
    assert receipt.status == ReceiptStatus.retried
    assert receipt.retried_at > now


def test_retry_created_receipt_payments():
    now = timezone.now()

    with freeze_time(now - datetime.timedelta(hours=3)):
        receipt1 = Receipt.objects.create()
    with freeze_time(now - datetime.timedelta(hours=18)):
        receipt2 = Receipt.objects.create()
    with freeze_time(now - datetime.timedelta(hours=28)):
        Receipt.objects.create()
    with freeze_time(now - datetime.timedelta(hours=3)):
        Receipt.objects.create(status='failed')
    Receipt.objects.create(status='received')

    with mock.patch.object(atol_create_receipt, 'delay') as task_mock:
        atol_retry_created_receipts()
        assert len(task_mock.mock_calls) == 2
        assert {call[0][0] for call in task_mock.call_args_list} == {receipt1.id, receipt2.id}


def test_retry_initiated_receipt_payments():
    now = timezone.now()

    receipt1 = Receipt.objects.create(status='initiated', initiated_at=now - datetime.timedelta(hours=1))
    receipt2 = Receipt.objects.create(status='initiated', initiated_at=now - datetime.timedelta(hours=18))
    Receipt.objects.create(status='initiated', initiated_at=now - datetime.timedelta(hours=28))
    Receipt.objects.create(status='failed', initiated_at=now - datetime.timedelta(hours=3))
    Receipt.objects.create(status='received')

    with mock.patch.object(atol_receive_receipt_report, 'delay') as task_mock:
        atol_retry_initiated_receipts()

    assert len(task_mock.mock_calls) == 2
    assert {call[0][0] for call in task_mock.call_args_list} == {receipt1.id, receipt2.id}



from datetime import datetime
from atol.utils import parse_receipt_datetime


def test_parse_receipt_datetime():
    assert parse_receipt_datetime('13.12.2017 18:55:19') == datetime(2017, 12, 13, 18, 55, 19)
    assert parse_receipt_datetime('2017.12.13') == datetime(2017, 12, 13, 0, 0)



from datetime import datetime
from uuid import uuid4
import responses
import pytest

from atol.core import AtolAPI
from atol.exceptions import AtolRecoverableError, AtolUnrecoverableError
from tests import ATOL_BASE_URL

ATOL_AUTH_CACHE_KEY = 'atol_auth_token:login'


@pytest.fixture
def caches():
    from django.core.cache import caches
    yield caches


@pytest.fixture
def set_atol_token(caches):
    def setter(token=None):
        token = token or str(uuid4().hex)[:12]
        caches['default'].set(ATOL_AUTH_CACHE_KEY, token)
        return token
    return setter


@responses.activate
def test_atol_create_receipt_workflow():
    uid = '973f3bef-1c39-40c9-abd0-33a91ab005ca'
    # get Token
    data = {
        'code': 1,
        'text': None,
        'token': '84a50b3a6207421aba46834d650b42a0'
    }
    url = ATOL_BASE_URL + '/getToken'
    responses.add(responses.POST, url, status=200, json=data)

    # sell
    data = {
        'uuid': uid,
        'timestamp': '13.07.2017 18:32:49',
        'status': 'wait',
        'error': None
    }
    url = ATOL_BASE_URL + '/ATOL-ProdTest-1/sell'
    responses.add(responses.POST, url, status=200, json=data)

    # report error
    data = {
        'uuid': uid,
        'timestamp': '13.07.2017 18:32:49',
        'status': 'wait',
        'error': {
            'code': 16,
            'text': 'Нет информации, попробуйте позднее',
            'type': 'system'
        },
        'payload': None
    }
    url = ATOL_BASE_URL + '/ATOL-ProdTest-1/report/' + uid
    responses.add(responses.GET, url, status=200, json=data)

    # report ok
    data = {
        'uuid': uid,
        'error': None,
        'status': 'done',
        'payload': {
            'total': 199.99,
            'fns_site': 'www.nalog.ru',
            'fn_number': '9999078900003780',
            'shift_number': 114,
            'receipt_datetime': '13.07.2017 18:32:00',
            'fiscal_receipt_number': 1412,
            'fiscal_document_number': 50066,
            'ecr_registration_number': '1029384756033729',
            'fiscal_document_attribute': 2649836604
        },
        'timestamp': '13.07.2017 18:32:50',
        'group_code': 'ATOL-ProdTest-1',
        'daemon_code': 'prod-agent-1',
        'device_code': 'KSR13.11-3-1',
        'callback_url': ''
    }
    url = ATOL_BASE_URL + '/ATOL-ProdTest-1/report/' + uid
    responses.add(responses.GET, url, status=200, json=data)

    # sell error
    data = {
        'uuid': uid,
        'timestamp': '13.07.2017 18:32:50',
        'status': 'fail',
        'error': {
            'code': 10,
            'text': 'В системе существует чек с external_id : ec0ce0c6-7a31-4f45-b94f-a1442be3bb9c '
                    'и group_code: ATOL-ProdTest-1',
            'type': 'system'
        }
    }
    url = ATOL_BASE_URL + '/ATOL-ProdTest-1/sell'
    responses.add(responses.POST, url, status=200, json=data)

    atol = AtolAPI()

    now = datetime(2017, 11, 22, 10, 47, 32)
    payment_uuid = str(uuid4())

    sell_params = dict(timestamp=now, transaction_uuid=payment_uuid,
                       purchase_name=u'Стандартная подписка на 1 месяц', purchase_price='199.99',
                       user_email='user@example.com', user_phone='+75551234567')

    receipt = atol.sell(**sell_params)

    assert receipt.uuid == '973f3bef-1c39-40c9-abd0-33a91ab005ca'
    assert receipt.data['status'] == 'wait'

    # report is not ready yet
    with pytest.raises(AtolRecoverableError):
        atol.report(receipt.uuid)

    report = atol.report(receipt.uuid)
    assert report.data['group_code'] == 'ATOL-ProdTest-1'
    assert report.data['status'] == 'done'
    assert report.data['payload']['total'] == 199.99

    # another celery worker somehow requested the same payment receipt
    double_receipt = atol.sell(**sell_params)
    assert double_receipt.uuid == '973f3bef-1c39-40c9-abd0-33a91ab005ca'
    assert double_receipt.data['status'] == 'fail'


@pytest.mark.parametrize('status,params', [
    (500, {'body': b''}),
    (302, {'body': b''}),
    (400, {'json': {'error': {'code': 1}}}),
    (400, {'json': {'error': {'code': 6}}}),
])
@responses.activate
def test_atol_sell_recoverable_errors(status, params, set_atol_token):
    atol = AtolAPI()
    set_atol_token('12345')

    sell_params = dict(timestamp=datetime.now(), transaction_uuid=str(uuid4()),
                       purchase_name=u'Стандартная подписка на 1 месяц', purchase_price='199.99',
                       user_email='user@example.com', user_phone='+75551234567')

    responses.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=status, **params)

    with pytest.raises(AtolRecoverableError):
        atol.sell(**sell_params)


@pytest.mark.parametrize('status,params', [
    (400, {'json': {'error': {'code': 3}}}),
    (400, {'json': {'error': {'code': 22}}}),
])
@responses.activate
def test_atol_sell_unrecoverable_errors(status, params, set_atol_token):
    atol = AtolAPI()
    set_atol_token('12345')

    sell_params = dict(timestamp=datetime.now(), transaction_uuid=str(uuid4()),
                       purchase_name=u'Стандартная подписка на 1 месяц', purchase_price='199.99',
                       user_email='user@example.com', user_phone='+75551234567')

    responses.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=status, **params)

    with pytest.raises(AtolUnrecoverableError):
        atol.sell(**sell_params)


def test_atol_sell_expired_token_is_renewed(set_atol_token, caches):
    atol = AtolAPI()
    set_atol_token('12345')

    assert caches['default'].get(ATOL_AUTH_CACHE_KEY) == '12345'

    receipt_uuid = str(uuid4())

    with responses.RequestsMock(assert_all_requests_are_fired=True) as resp_mock:
        resp_mock.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=401)
        resp_mock.add(responses.POST, ATOL_BASE_URL + '/getToken', status=200,
                      json={'code': 0, 'token': 'foobar'})
        resp_mock.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=200,
                      json={'uuid': receipt_uuid})

        sell_params = dict(timestamp=datetime.now(), transaction_uuid=str(uuid4()),
                           purchase_name=u'Стандартная подписка на 1 месяц', purchase_price='199.99',
                           user_email='user@example.com', user_phone='+75551234567')

        receipt = atol.sell(**sell_params)
        assert receipt.uuid == receipt_uuid
        assert resp_mock.calls[0].request.url.endswith('?tokenid=12345')

    # token is updated
    assert caches['default'].get(ATOL_AUTH_CACHE_KEY) == 'foobar'

    with responses.RequestsMock(assert_all_requests_are_fired=True) as resp_mock:
        anoher_receipt_uuid = str(uuid4())
        resp_mock.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=200,
                      json={'uuid': anoher_receipt_uuid})

        sell_params = dict(timestamp=datetime.now(), transaction_uuid=str(uuid4()),
                           purchase_name=u'Стандартная подписка на 1 месяц', purchase_price='199.99',
                           user_email='user@example.com', user_phone='+75551234567')

        receipt = atol.sell(**sell_params)
        assert receipt.uuid == anoher_receipt_uuid
        assert resp_mock.calls[0].request.url.endswith('?tokenid=foobar')


def test_atol_sell_expired_token_is_failed_to_renew(set_atol_token):
    atol = AtolAPI()
    set_atol_token()

    with responses.RequestsMock(assert_all_requests_are_fired=True) as resp_mock:
        resp_mock.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/sell', status=401)
        resp_mock.add(responses.POST, ATOL_BASE_URL + '/getToken', status=400, json={'code': 19})

        sell_params = dict(timestamp=datetime.now(), transaction_uuid=str(uuid4()),
                           purchase_name=u'Стандартная подписка на 1 месяц', purchase_price='199.99',
                           user_email='user@example.com', user_phone='+75551234567')

        with pytest.raises(AtolRecoverableError):
            atol.sell(**sell_params)


@pytest.mark.parametrize('status,params', [
    (500, {'body': b''}),
    (302, {'body': b''}),
    (400, {'json': {'error': {'code': 7}}}),
    (400, {'json': {'error': {'code': 12}}}),
    (400, {'json': {'error': {'code': 16}}}),
])
@responses.activate
def test_atol_report_recoverable_errors(status, params, set_atol_token):
    atol = AtolAPI()
    set_atol_token('12345')
    payment_uuid = str(uuid4())

    responses.add(responses.POST, ATOL_BASE_URL + '/ATOL-ProdTest-1/report/%s' % payment_uuid,
                  status=status, **params)

    with pytest.raises(AtolRecoverableError):
        atol.report(payment_uuid)


@pytest.mark.parametrize('status,params', [
    (400, {'json': {'error': {'code': 3}}}),
    (400, {'json': {'error': {'code': 15}}}),
])
@responses.activate
def test_atol_report_unrecoverable_errors(status, params, set_atol_token):
    atol = AtolAPI()
    set_atol_token('12345')
    payment_uuid = str(uuid4())

    responses.add(responses.GET, ATOL_BASE_URL + '/ATOL-ProdTest-1/report/%s' % payment_uuid,
                  status=status, **params)

    with pytest.raises(AtolUnrecoverableError):
        atol.report(payment_uuid)



import logging
import datetime
from dateutil.parser import parse as parse_date

logger = logging.getLogger(__name__)


def parse_receipt_datetime(date_str):
    try:
        return datetime.datetime.strptime(date_str, '%d.%m.%Y %H:%M:%S')
    except Exception:
        logger.warning('unexpected date format in receipt, %s', date_str)
        return parse_date(date_str)



import logging
import requests
from collections import namedtuple
from dateutil.parser import parse as parse_date

from django.conf import settings
from django.core.cache import cache

from atol import exceptions

logger = logging.getLogger(__name__)

NewReceipt = namedtuple('NewReceipt', ['uuid', 'data'])
ReceiptReport = namedtuple('ReceiptReport', ['uuid', 'data'])


class AtolAPI(object):
    base_url = 'https://online.atol.ru/possystem/v3'
    request_timeout = 5

    def _obtain_new_token(self):
        """
        Obtain new access token using the login-password credentials pair.
        If failed to obtain token, raise an exception.
        """
        response_data = self._request('post', 'getToken', json={
            'login': settings.RECEIPTS_ATOL_LOGIN,
            'pass': settings.RECEIPTS_ATOL_PASSWORD,
        })
        # all codes other than 0 (new token) and 1 (existing token) are considered errors
        if response_data.get('code') not in (0, 1):
            raise exceptions.AtolAuthTokenException()

        auth_token = response_data['token']
        logger.info('successfully obtained fresh auth token "%s" for login "%s"',
                    auth_token, settings.RECEIPTS_ATOL_LOGIN)
        return auth_token

    def _get_auth_token(self, force_renew=False):
        """
        Obtain the authentication token obtained earlier and put in cache.
        In case the cache yields nothing, or the token has expired (401),
        obtain a new token then put in cache.
        """
        cache_key = 'atol_auth_token:{login}'.format(login=settings.RECEIPTS_ATOL_LOGIN)
        auth_token = cache.get(cache_key)

        # obtain and store fresh auth token
        if not auth_token or force_renew:
            auth_token = self._obtain_new_token()
            # cache the key forever without a ttl
            cache.set(cache_key, auth_token)
        else:
            logger.debug('successfully obtained auth token "%s" for login "%s" from cache',
                         auth_token, settings.RECEIPTS_ATOL_LOGIN)

        return auth_token

    def request(self, method, endpoint, json=None):
        """
        Make a request to atol api endpoint using cached access token.

        If endpoint yields a 401 error, obtain a new token then try the request again.

        :param method: HTTP method
        :param endpoint: Name of atol endpoint (e.g. sell, report, you name it)
        :param json: json-ready request data (normally this would be a dict)

        The final url will assume the following form:
            https://online.atol.ru/possystem/v3/MyCompany_MyShop/sell?tokenid=d8c7021934fg4f2384ebf6b72624bbbf
        """
        auth_token = self._get_auth_token()

        params = {
            'tokenid': auth_token
        }

        # signed requests contain group codes in front of the endpoint name
        endpoint = '{group_code}/{endpoint}'.format(group_code=settings.RECEIPTS_ATOL_GROUP_CODE,
                                                    endpoint=endpoint)

        try:
            return self._request(method, endpoint, params=params, json=json)
        except exceptions.AtolAuthTokenException:
            # token must have expired, try new one
            logger.info('trying new token for request "%s" to endpoint %s with params=%s json=%s token=%s',
                        method, endpoint, params, json, auth_token)
            params.update({'tokenid': self._get_auth_token(force_renew=True)})
            return self._request(method, endpoint, params=params, json=json)

    def _request(self, method, endpoint, params=None, headers=None, json=None):
        params = params or {}
        headers = headers or {}
        headers.setdefault('Content-Type', 'application/json')

        url = '{base_url}/{endpoint}'.format(base_url=self.base_url.rstrip('/'),
                                             endpoint=endpoint)

        logger.info('about to %s %s with headers=%s, params=%s json=%s', method, url, headers, params, json)

        try:
            response = requests.request(method, url, params=params, json=json,
                                        headers=headers, timeout=self.request_timeout)
        except Exception as exc:
            logger.warning('failed to request %s %s with headers=%s, params=%s json=%s due to %s',
                           method, url, headers, params, json, exc,
                           exc_info=True,
                           extra={'data': {'json': json, 'params': params}})
            raise exceptions.AtolRequestException()

        # error codes other than 2xx, 400, 401 are considered unexpected and yield an exception
        if response.status_code not in (200, 201, 400, 401):
            try:
                json_response = response.json()
            except Exception:
                json_response = None
            logger.warning('request %s %s with headers=%s, params=%s json=%s failed with status code %s: %s',
                           method, url, headers, params, json, response.status_code, json_response,
                           extra={'data': {'json_request': json,
                                           'content': response.content,
                                           'json_response': json_response}})
            raise exceptions.AtolRequestException()

        # 401 should be handled separately by the calling code
        if response.status_code == 401:
            logger.info('authentication failed for request %s %s with headers=%s, params=%s json=%s',
                        method, url, headers, params, json, extra={'data': {'content': response.content}})
            raise exceptions.AtolAuthTokenException()

        try:
            response_data = response.json()
        except Exception as exc:
            logger.warning('unable to parse json response due to %s', exc, exc_info=True,
                           extra={'data': {'content': response.content}})
            raise exceptions.AtolRequestException()

        if response_data.get('error'):
            logger.warning('received error response from atol url %s: %s',
                           url, response_data['error'],
                           extra={'data': {'json': json, 'params': params}})
            raise exceptions.AtolClientRequestException(response=response,
                                                        response_data=response_data,
                                                        error_data=response_data['error'])

        return response_data

    def sell(self, **params):
        """
        Register a new receipt for given payment details on the atol side.
        Receive receipt uuid for the created receipt.

        :param timestamp: Payment datetime
        :param transaction_uuid: Unique payment id (potentically across all organization projects' payments.
                                 uuid4 should do fine.
        :param purchase_name: Human readable name of the purchased product
        :param purchase_price: The amount in roubles the user was billed with
        :param user_email: User supplied email
        :param user_phone: User supplied phone (may or may not start with +7)
        """
        user_email = params.get('user_email')
        user_phone = params.get('user_phone')
        # receipt must contain either of the two
        if not (user_email or user_phone):
            raise exceptions.AtolPrepRequestException()

        purchase_price = params['purchase_price']
        # convert decimals and strings to float, because atol does not accept those types
        if not isinstance(purchase_price, int):
            purchase_price = float(purchase_price)

        timestamp = params['timestamp']
        if isinstance(timestamp, str):
            timestamp = parse_date(timestamp)

        request_data = {
            'external_id': params['transaction_uuid'],
            'timestamp': timestamp.strftime('%d.%m.%Y %H:%M:%S'),
            'receipt': {
                # user supplied details
                'attributes': {
                    'email': user_email or u'',
                    'phone': user_phone or u'',
                },
                'items': [{
                    'name': params['purchase_name'],
                    'price': purchase_price,
                    'quantity': 1,
                    'sum': purchase_price,
                    'tax': settings.RECEIPTS_ATOL_TAX_NAME,
                }],
                'payments': [{
                    'sum': purchase_price,
                    'type': 1,
                }],
                'total': purchase_price,
            },
            'service': {
                'inn': settings.RECEIPTS_ATOL_INN,
                'callback_url': settings.RECEIPTS_ATOL_CALLBACK_URL or u'',
                'payment_address': settings.RECEIPTS_ATOL_PAYMENT_ADDRESS,
            }
        }

        try:
            response_data = self.request('post', 'sell', json=request_data)
        # check for recoverable errors
        except exceptions.AtolClientRequestException as exc:
            logger.info('sell request with json %s failed with code %s', request_data, exc.error_data['code'])
            if exc.error_data['code'] in (1, 4, 5, 6):
                raise exceptions.AtolRecoverableError()
            if exc.error_data['code'] == 10:
                logger.info('sell request with json %s already accepted; uuid: %s',
                            request_data, exc.response_data['uuid'])
                return NewReceipt(uuid=exc.response_data['uuid'], data=exc.response_data)
            # the rest of the errors are not recoverable
            raise exceptions.AtolUnrecoverableError()
        except Exception as exc:
            logger.warning('sell request with json %s failed due to %s', request_data, exc, exc_info=True)
            raise exceptions.AtolRecoverableError()

        return NewReceipt(uuid=response_data['uuid'], data=response_data)

    def report(self, receipt_uuid):
        """
        The receipt may not yet be processed by the time of the request,
        the calling code should try this method again later.

        :param receipt_uuid: Receipt identifier previously returned by atol
        """
        try:
            response_data = self.request('get', 'report/{uuid}'.format(uuid=receipt_uuid))
        # check for recoverable errors
        except exceptions.AtolClientRequestException as exc:
            logger.info('report request for receipt %s failed with code %s', receipt_uuid, exc.error_data['code'])
            if exc.error_data['code'] in (7, 9, 12, 13, 14, 16):
                raise exceptions.AtolRecoverableError()
            if exc.error_data['code'] == 1:
                logger.info('report request for receipt %s was not processed: %s; '
                            'Must repeat the request with a new unique value <external_id>',
                            receipt_uuid, exc.response_data.get('text'))
                raise exceptions.AtolReceiptNotProcessed(exc.response_data.get('text'))
            # the rest of the errors are not recoverable
            raise exceptions.AtolUnrecoverableError()
        except Exception as exc:
            logger.info('report request for receipt %s failed due to %s', receipt_uuid, exc)
            raise exceptions.AtolRecoverableError()

        return ReceiptReport(uuid=receipt_uuid, data=response_data)



import logging
import shortuuid

from django.conf import settings
from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.utils.encoding import force_bytes
from django.views.generic import RedirectView
from django.utils.translation import ugettext_lazy as _

from atol.models import Receipt
from atol.exceptions import MissingReceipt
from atol.utils import parse_receipt_datetime

logger = logging.getLogger(__name__)


class ReceiptView(RedirectView):
    """
    Anonymous view as a proxy to OFD provider.
    - link is shorter for sms usage (use internal uuid, so that we have stable link for our payments)
    - ability to change ofd at any point
    - links will not break even if ofd-1 changes URL scheme
    - easy way to monitor feature usage via logs

    Swallows missing receipts and malformed data to show only 404 (user-friendly)
    """

    def get(self, request, *args, **kwargs):
        try:
            return super(ReceiptView, self).get(request, *args, **kwargs)
        except MissingReceipt:
            return HttpResponseNotFound(content=force_bytes(_('Чек не найден')))
        except (KeyError, TypeError, ValueError) as exc:
            logger.error('invalid receipt format: %s', exc, exc_info=True)
            return HttpResponseNotFound(content=force_bytes(_('Чек не найден')))  # do not face 500 to user

    def get_redirect_url(self, *args, **kwargs):
        uuid = shortuuid.decode(kwargs['short_uuid'])
        receipt = get_object_or_404(Receipt, internal_uuid=uuid)

        if not receipt.content:
            logger.warning('access receipt before backend processed receipt, suspicious')
            raise MissingReceipt()

        payload = receipt.content['payload']
        receipt_dt = parse_receipt_datetime(payload['receipt_datetime'])

        return settings.RECEIPTS_OFD_URL_TEMPLATE.format(
            t='{:%Y%m%dT%H%M%S}'.format(receipt_dt),
            s=payload['total'],
            fn=payload['fn_number'],
            fd=payload['fiscal_document_number'],
            fp=payload['fiscal_document_attribute'],
            n=payload['fiscal_receipt_number'],
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-11-27 20:43
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Receipt',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('internal_uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Дата создания чека')),
                ('initiated_at', models.DateTimeField(blank=True, null=True, verbose_name='Дата инициализации чека в системе оператора')),
                ('received_at', models.DateTimeField(blank=True, null=True, verbose_name='Дата получения чека от оператора')),
                ('failed_at', models.DateTimeField(blank=True, null=True, verbose_name='Дата ошибки')),
                ('status', models.CharField(choices=[('created', 'Ожидает инициации в системе оператора'), ('initiated', 'Иницирован в системе оператора'), ('received', 'Получен от оператора'), ('no_email_phone', 'Отсутствует email/phone'), ('failed', 'Ошибка')], default='created', max_length=16, verbose_name='Статус чека')),
                ('uuid', models.TextField(editable=False, help_text='Идентификатор чека платежа в системе оператора', null=True, verbose_name='Идентификатор чека')),
                ('content', django.contrib.postgres.fields.jsonb.JSONField(editable=False, null=True, verbose_name='Содержимое чека')),
                ('user_email', models.CharField(max_length=254, null=True, verbose_name='Email пользователя')),
                ('user_phone', models.CharField(max_length=32, null=True, verbose_name='Телефон пользователя')),
                ('purchase_price', models.DecimalField(decimal_places=2, max_digits=8, null=True, verbose_name='Цена покупки')),
                ('purchase_name', models.TextField(null=True, verbose_name='Наименование покупки')),
            ],
            options={
                'verbose_name': 'Чек Атола',
                'verbose_name_plural': 'Чеки Атола',
                'ordering': ['id'],
            },
        ),
    ]



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = 'super-secret-string'
DEBUG = False


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'atol',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'tests.test_app.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


# Database
# https://docs.djangoproject.com/en/1.11/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'atol'
    }
}


# Caches
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'unique-snowflake',
    }
}


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/

CELERY_ALWAYS_EAGER = True
CELERY_IGNORE_RESULT = True

RECEIPTS_ATOL_LOGIN = 'login'
RECEIPTS_ATOL_PASSWORD = 'secret'
RECEIPTS_ATOL_GROUP_CODE = 'ATOL-ProdTest-1'
RECEIPTS_ATOL_TAX_NAME = 'vat18'
RECEIPTS_ATOL_INN = '112233445573'
RECEIPTS_ATOL_CALLBACK_URL = None
RECEIPTS_ATOL_PAYMENT_ADDRESS = 'г. Москва, ул. Оранжевая, д.22 к.11'
RECEIPTS_OFD_URL_TEMPLATE = u'https://lk.platformaofd.ru/web/noauth/cheque?fn={fn}&fp={fp}'



#!/usr/bin/env python
import os
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


with open('README.rst') as readme_file:
    readme = readme_file.read()


with open('CHANGELOG.rst') as history_file:
    history = history_file.read()


requirements = [
    'requests>=2.10.0',
    'python-dateutil>=2.5.3',
    'django-model-utils>=3.0.0',
    'celery>=3.1.25',
    'shortuuid>=0.5.0',
]

setup(
    name='django-atol',
    version=get_version('atol'),
    description='Django integration with ATOL online',
    long_description=readme + '\n\n' + history,
    author='MyBook',
    author_email='dev@mybook.ru',
    url='https://github.com/MyBook/django-atol',
    packages=[
        'atol',
        'atol.migrations'
    ],
    package_dir={'atol': 'atol'},
    include_package_data=True,
    install_requires=requirements,
    license='BSD',
    zip_safe=False,
    keywords='atol',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
)



"""
Signals:

receipt_initiated:
    Called at the time of successful receipt registration in Atol

receipt_failed:
    Called if an error occurred:
    - There is not enough data to register a receipt in Atol
    - Error registering receipt
    - Error of receipt processing

receipt_received:
    Called at the moment of Atol's response on successful processing of the receipt

"""
from django.dispatch import Signal

receipt_initiated = Signal(providing_args=['receipt'])
receipt_failed = Signal(providing_args=['receipt'])
receipt_received = Signal(providing_args=['receipt'])



class NoEmailAndPhoneError(Exception):
    pass


class MissingReceipt(Exception):
    pass


class AtolException(Exception):
    pass


class AtolPrepRequestException(AtolException):
    """Raised client side due to various validation errors prior to a request"""
    pass


class AtolRequestException(AtolException):
    """Raised when atol request fails in the middle, or the endpoint returns unexpected response"""
    pass


class AtolClientRequestException(AtolException):
    """Raised when received an expected error code from atol"""

    def __init__(self, *args, **kwargs):
        self.response = kwargs.pop('response', None)
        self.response_data = kwargs.pop('response_data')
        self.error_data = kwargs.pop('error_data')
        super(AtolClientRequestException, self).__init__(*args, **kwargs)


class AtolAuthTokenException(AtolException):
    """Raised when atol api endpoint returns 401"""
    pass


class AtolUnrecoverableError(AtolException):
    """Raised when atol responds with an error
       that is generally unrecoverable and requires manual maintenance"""
    pass


class AtolRecoverableError(AtolException):
    """Raised when atol responds with an error
       that is normally recoverable and can be retried with an extra request"""
    pass


class AtolReceiptNotProcessed(AtolException):
    """Raised when atol responds with an error
        that is requires a repeated request to register a check

        Call if:
        1. The number of KKT in the group does not correspond to the speed of receipt of checks.
            It is necessary to connect additional KKT or reduce the speed of receipt of checks.
        2. All cash desks of the group are off or inactive. Please contact your administrator.
    """
    pass
